package org.expeditee.network;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

import org.expeditee.gui.AttributeValuePair;
import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameIO;
import org.expeditee.gui.MessageBay;
import org.expeditee.io.ExpReader;
import org.expeditee.io.ExpWriter;
import org.expeditee.io.FrameReader;
import org.expeditee.io.FrameWriter;
import org.expeditee.items.Text;

public class FrameShare {

	public static boolean disableNetworking = false;
	
	private static final int SOCKET_TIMEOUT = 1000;

	private static Collection<DefaultServer> _servers = new LinkedList<DefaultServer>();

	private static FrameShare _theSession;

	private Map<String, Peer> _peers;

	private int _port = Peer.DEFAULT_PORT;


        private void startServers() throws IOException
        {
	    System.err.println("Starting Expeditee Server on port " + _port);
	    _servers.add(new FrameServer(_port));
	    _servers.add(new FrameSaver(_port));
	    _servers.add(new MessageReciever(_port));
	    _servers.add(new InfServer(_port));
	    _servers.add(new InfUpdate(_port));

	}

	private FrameShare() {
		_peers = new HashMap<String, Peer>();
	}

	private FrameShare(int port) {
	        this();
		_port = port;

		try {
		    startServers();
		    
		} catch (Exception e) {
		    e.printStackTrace();
		    System.err.println("Could not start servers [port: "
				       + port + "]");
		}

		try {
		    for (DefaultServer server : _servers) {
			server.start();
		    }
		} catch (Exception e) {
		    System.err.println("Error in server startup");
		}
	}


 	private FrameShare(Frame settingsFrame) 
        {
	        this();

		// Set the settings
		for (Text item : settingsFrame.getBodyTextItems(false)) {

			AttributeValuePair avp = new AttributeValuePair(item.getText());
			if (avp.isAnnotation())
				continue;

			String attribute = avp.getAttributeOrValue().toLowerCase();

			if (attribute.equals("server")) {
				try {
					if (avp.hasPair()) {
						_port = avp.getIntegerValue();
					}
					MessageBay.displayMessage("Starting Expeditee Server on port: " + _port);
					
					startServers();
				} catch (Exception e) {
					e.printStackTrace();
					MessageBay.errorMessage("Could not start servers ["
							+ avp.toString() + "]");
				}
				continue;
			}

			if (!avp.hasPair())
				continue;

			try {
			    Peer peer = new Peer(avp);
			    _peers.put(attribute, peer);
			} catch (UnknownHostException e) {
				MessageBay.errorMessage("Could not locate peer ["
						+ avp.toString() + "]");
			}
		}

		try {
		    for (DefaultServer server : _servers) {
				server.start();
		    }
		} catch (Exception e) {
			MessageBay.errorMessage("Error in PeerToPeer setup");
		}
	}

	public void finalise() {
	    System.err.println("Closing servers");

		for (DefaultServer server : _servers)
			server.close();
	}

	public static FrameShare getInstance() {
		return _theSession;
	}

	/**
	 * TODO check each peer on a different thread.
	 * 
	 * @param frameName
	 * @return
	 */
	public Frame loadFrame(String frameName, String peerName) {
		String result = null;

		try {
			// get a datagram socket
			DatagramSocket socket = new DatagramSocket();
			socket.setSoTimeout(SOCKET_TIMEOUT * 2);
			if (peerName == null) {
				for (Peer peer : _peers.values()) {
					try {
						result = getFrameContents(frameName, socket, peer);
						if(result == null || result.length() == 0) {
							continue;
						}
						peerName = peer.getName();
						break;
					} catch (Exception e) {
						e.printStackTrace();
					}
				}
			} else {
				try {
					Peer peer = _peers.get(peerName.toLowerCase());
					if (peer != null) {
						result = getFrameContents(frameName, socket, peer);
					}
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			socket.close();
		} catch (Exception e) {
		    e.printStackTrace();
		}

		if (result == null || result.length() == 0)
			return null;

		// Now read the frame from the file contents
		FrameReader reader = new ExpReader(frameName);
		Frame frame = null;
		try {
			frame = reader.readFrame(new BufferedReader(
					new StringReader(result)));
			// Set the path for the frame to indicate it is NOT a local frame
			// This allows the frame to be saved in the correct location
			frame.setLocal(false);
			frame.setPath(peerName);
		} catch (IOException e) {
			e.printStackTrace();
		}

		if (frame == null) {
			MessageBay.errorMessage("Error: " + frameName
					+ " could not be successfully loaded.");
			return null;
		}

		return frame;
	}

	/**
	 * @param frameName
	 * @param socket
	 * @param peer
	 * @return
	 * @throws IOException
	 */
	private String getFrameContents(String frameName, DatagramSocket socket, Peer peer) throws IOException {
		byte[] nameBuf = frameName.getBytes();
		byte[] buf = new byte[FrameServer.MAX_PACKET_LENGTH];
		// send request for a frame
		DatagramPacket packet = new DatagramPacket(nameBuf,
				nameBuf.length, peer.getAddress(), peer
						.getPort());
		socket.send(packet);

		// get response
		packet = new DatagramPacket(buf, buf.length);
		socket.receive(packet);

		// store frame contents
		return new String(packet.getData(), 0, packet
				.getLength());
	}

	public boolean sendMessage(String message, String peerName) {
		Peer peer = _peers.get(peerName.toLowerCase());

		if (peer == null) {
			return false;
		}

		try {
			// get a datagram socket
			DatagramSocket socket = new DatagramSocket(_port - 1);
			socket.setSoTimeout(SOCKET_TIMEOUT);

			// message = peerName + " says " + message;
			byte[] contentsBuf = message.getBytes();

			try {
				// send save request
				DatagramPacket packet = new DatagramPacket(contentsBuf,
						contentsBuf.length, peer.getAddress(), peer.getPort()
								+ MessageReciever.OFFSET);
				socket.send(packet);
			} catch (Exception e) {
				e.printStackTrace();
			}
			socket.close();

		} catch (IOException e) {
			e.printStackTrace();
		}
		return true;
	}

	public String saveFrame(Frame toSave) {

		FrameIO.setSavedProperties(toSave);

		Peer peer = _peers.get(toSave.getPath().toLowerCase());
		String fileContents = "";
		// Now read the frame from the file contents
		FrameWriter writer = new ExpWriter();
		try {
			// String tempName = new Date().toString() + ".exp";
			// writer.setOutputLocation(tempName);
			// Write out the file to a StringBuffer
			StringWriter sw = new StringWriter();
			// First write out the name of the frame
			sw.write(toSave.getName() + "\n");
			// Then the version
			sw.write(toSave.getVersion() + "\n");
			// Write out the rest of the frame
			writer.writeFrame(toSave, sw);
			// Now send the packet
			fileContents = sw.getBuffer().toString();
			byte[] contentsBuf = fileContents.getBytes();

			// get a datagram socket
			DatagramSocket socket = new DatagramSocket(_port - 2);
			socket.setSoTimeout(SOCKET_TIMEOUT);

			try {
				// send save request
				DatagramPacket packet = new DatagramPacket(contentsBuf,
						contentsBuf.length, peer.getAddress(), peer.getPort()
								+ FrameSaver.OFFSET);
				socket.send(packet);
			} catch (Exception e) {
				e.printStackTrace();
			}
			socket.close();

		} catch (IOException e) {
			e.printStackTrace();
		}
		toSave.setSaved();
		return fileContents;
	}

	public String getPeerName(int port, InetAddress address) {
		for (Peer p : _peers.values()) {
			if (p.getPort() == port && p.getAddress().equals(address))
				return p.getName();
		}
		return null;
	}

	public int getInfNumber(String peerName, String frameset, boolean update)
			throws IOException {
		int result = -1;
		// get a datagram socket
		DatagramSocket socket = null;
		try {
			socket = new DatagramSocket(_port - 3);
			socket.setSoTimeout(SOCKET_TIMEOUT);

			byte[] contentsBuf = frameset.getBytes();
			Peer peer = _peers.get(peerName.toLowerCase());
			try {
				// send inf request
				DatagramPacket packet = new DatagramPacket(
						contentsBuf,
						contentsBuf.length,
						peer.getAddress(),
						peer.getPort()
								+ (update ? InfUpdate.OFFSET : InfServer.OFFSET));
				socket.send(packet);

				byte[] buf = new byte[100];
				// get response
				packet = new DatagramPacket(buf, buf.length);
				socket.receive(packet);

				// store frame contents
				result = Integer.parseInt(new String(packet.getData(), 0,
						packet.getLength()));
				peerName = peer.getName();

			} catch (Exception e) {
				e.printStackTrace();
			}
			socket.close();
		} catch (SocketException e1) {
			e1.printStackTrace();
		}

		return result;
	}

	public static void init(Frame settingsFrame) {
		if (disableNetworking || settingsFrame == null)
			return;

		if (_theSession == null)
			_theSession = new FrameShare(settingsFrame);
	}

	public static void init(int port) {

	        if (_theSession == null) {
			_theSession = new FrameShare(port);
		}
	}


	/**
	 * Start a server running on the port number supplied on the command line
	 * 
	 * @param args
	 */
	public static void main(String[] args) {

	    if (args.length != 1) {

		// Work out the 'program name' (i.e. this class), but done in a general 
		//   way in case class name is changed at a later date
		StackTraceElement[] stack = Thread.currentThread ().getStackTrace ();
		StackTraceElement main = stack[stack.length - 1];
		String mainClass = main.getClassName ();

		System.err.println("Usage: java " + mainClass + " port-number");
		System.exit(1);
	    }
	    
	    String port_str = args[0];

	    try {
		int port = Integer.parseInt(port_str);

		init(port);
	    }
	    catch (Exception e) {
		e.printStackTrace();
		System.exit(2);
	    }

	}

}
